%************************************************************************/
%* The Bayesian Change Point algorithm - A program to caluclate the     */
%* posterior probability of a change point in a time series.            */
%*                                                                      */
%* Please acknowledge the program author on any publication of          */
%* scientific results based in part on use of the program and           */
%* cite the following article in which the program was described.       */
%*                                                                      */
%* E. Ruggieri.  "A Bayesian Approach to Detecting Change Points        */
%* in Climatic Records," International Journal of Climatology           */
%*                                                                      */
%* Program Author: Eric Ruggieri                                        */
%* Duquesne University                                                  */
%* Pittsburgh, PA 15282                                                 */
%* Email:  ruggierie@duq.edu                                            */
%*                                                                      */
%* Copyright (C) 2013  Duquesne University                              */
%*                                                                      */
%* The Bayesian Change Point algorithn is free software: you can        */
%* redistribute it and/or modify it under the terms of the GNU General  */
%* Public License as published by the Free Software Foundation, either  */
%* version 3 of the License, or (at your option) any later version.     */
%*                                                                      */
%* The Bayesian Change Point algorithm is distributed in the hope that  */
%* it will be useful, but WITHOUT ANY WARRANTY; without even the        */
%* implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR      */
%* PURPOSE.  See the GNU General Public License for more details.       */
%*                                                                      */
%* You should have received a copy of the GNU General Public License    */
%* along with Bayesian Change Point.  If not, see                       */
%* <http://www.gnu.org/licenses/> or write to the                       */
%* Free Software Foundation, Inc.                                       */
%* 51 Franklin Street Fifth Floor                                       */
%* Boston, MA 02110-1301, USA.                                          */
%************************************************************************/

% Outline of the Bayesian Change Point algorithm:
% 1) Load the Data  
% 2) Define the Parameter Values
% 3) Calculate the Probability Density of the Data for Each Sub-Interval
% 4) Forward Recursion [Dynamic Programming]
% 5) Stochastic Backtrace via Bayes Rule
%       a) Sample a Number of Change Points
%       b) Sample the Location of the Change Points
%       c) Sample the Regression Parameters Between Adjacent Change Points
% 6) Plot the Results
%
% Description of Parameters, input, and output can be found in accompanying 
% ReadMe file
%

clear;
%************(1)Load the Data*****************************

load LR04stack.txt;
LR04stack=LR04stack(:,1:2);
x=LR04stack(:,1);

%A Linear Model with non-polynomial terms.  This method used a 'design
%matrix".  For more information, Google matlab
%detrending and click on the first link: mathworks.com
t=LR04stack(:,1);
y=LR04stack(:,2);
X = [ones(size(t))  exp(-t/1000)  t.*exp(-t/1000)];
a=X\y;
Y = LR04stack(:,2) - (a(1) + a(2)*exp(-t/1000) +a(3)*t.*exp(-t/1000));

clear t y X;
disp('Data Detrended');

w1=2*pi/23;
w2=2*pi/41;
w3=2*pi/100;

%*************(2)Define the Parameter Values**************

N=length(Y);    % Number of data points
k_max=20;        % Maximum number of change points
d_min=15;       % Minimum distance between adjacent change points
      
k_0=0.01;           % Hyperparameter for the prior on the regression coefficients
v_0=10; sig_0=0.30;  % Hyperparameter for scaled inverse chi-square prior on the variance

X = [zeros(N,1)+1 cos(w1*x) cos(w2*x) cos(w3*x) sin(w1*x) sin(w2*x) sin(w3*x)];


[~, m] =size(X);             % r = total # data points (N), m = # regressors
beta0= zeros(m,1);          % Mean of multivariate normal prior on regression coefficients
num_samp=500;               % Number of sampled solutions

parameters = [d_min k_0 v_0 sig_0 k_max num_samp];

%******(3)Calculate the Probability Density of Data for Each Sub-Interval*********
Py=zeros(N,N)-Inf;      % Matrix containing the probability of each substring of the data
I=eye(m);               % cxc identity matrix

for i=1:N
    for j=i+d_min-1:N
        
        %Calculate beta_hat
        XTy=X(i:j,:)'*Y(i:j); 
        J=k_0*I+X(i:j,:)'*X(i:j,:); J=inv(J);
        beta_hat=J*(k_0*beta0+XTy);
        
        %Calculate v_n and s_n
        a=j-i+1;    % 'a' is the number of data points.... not the distance between them, 
                    % which becomes relevant if data points are not equally spaced
        v_n=v_0+a;  % a is the number of data points in the segment, see above for definition
        
        s_n = v_0*sig_0 + k_0*(beta0-beta_hat)'*(beta0-beta_hat) + (Y(i:j,:)-X(i:j,:)*beta_hat)'*(Y(i:j,:)-X(i:j,:)*beta_hat);
        
        % Calculate the Probability Density of the Data - Equation (2) - stored in log form 
        Py(i,j) = v_0*log(sig_0*v_0/2)/2 + gammaln(v_n/2) +m*log(k_0)/2  ...
            - log(gamma(v_0/2)) - v_n*log(s_n/2)/2 - a*log(2*pi)/2 + log(det(J))/2 ;
     end
end

%****************(4) Build Partition Function *************************
P =partition_fn(Py, k_max, N);
       
%****************(5) Stochastic Backtrace via Bayes Rule****************

k= P(:,N);
for i=1:k_max  % Equation (6)
    if(N-(i+1)*d_min+i >= i)
        k(i) = k(i) + log(0.5) - log(k_max) - log(nCk(N-(i+1)*d_min+i,i));
    end
end
k=[Py(1,N)+ log(0.5); k];

total_k = log(sum(exp(k)));
k(:)=exp(k(:)-total_k);     % Normalize the vector - Equation (7)



[samp_holder, chgpt_loc] = plot_results_pruned(parameters, x, X, Y, Py, P, k);
