function [distanceF, distanceO] = distance_function (samp_holderF, samp_holderO)
% This function calculates the distance between two different change point
% solutions.  If the solutions are the same length, then the distance is
% the difference between the change point positions.  If the solutions
% contain a different number of change points, then the sequences are first
% aligned (Needleman-Wunsch) before calculating the difference.  Gap
% penalty is set at the length of the sequence.

% A K-Nearest Neighbor distance function is being used to approximate
% Kullback-Leibarber divergence (since we are working in high-dimensional
% space).  See "Divergence Estimation for Multidimensional Densities Via
% k-Nearest-NEighbor Distances" by Wang, Kulkarni, and Verdu

% Note: The way the function is written, we are calculating the divergence
% of "F" from "O", where "O" is meant to be the original /true
% distribution.

num_samp = max(size(samp_holderF)); 
N = 3970;   % Length of well-log data set
K = 50;     % number of nearest neighbors

distanceO = zeros(1,num_samp);      % Holds distance to Kth nearest neighbor to "O" sequences
distanceF = zeros(1,num_samp);      % Given a solution in "O", the distance to the Kth nearest neighbor among the "F" sequences

nF = zeros(num_samp,1);         % Number of change points in each solution
nO = zeros(num_samp,1);
d = zeros(num_samp,1);          % holds the dimension of the solution

n = min(size(samp_holderF));    % Equivalent to kmax

for i = 1:num_samp
    nF(i) = min([find(samp_holderF(i,:) ==0,1)-1 n]);   % gives number of change points in each solution
    nO(i) = min([find(samp_holderO(i,:) ==0,1)-1 n]);
end

% Our goal is to find the distance of a change point solution to its K
% nearest neighbors in both distributions.  The FAST distribution is the
% approximation of the true distribution, ORIGINAL
for i = 1:num_samp
    % We have to start by finding the set of 'near' neighbors to analyze
    doneF = 0; doneO = 0; gapF = 0; gapO = 0;
    
    while (~doneF)
        
        enough = 0;
        while (~enough)
            testF = find(nF >= nO(i)-gapF & nF <= nO(i)+gapF);       % These are all the samples of the same length
            if (length(testF) > K)
                enough = 1;                             % We potentially have enough samples
            else
                gapF = gapF + 1;                          % Otherwise let's expand the # chgpts we will consider
            end
        end
        
        distF = zeros(length(testF), 1);    % Holds distance to each possible change point solution identified in set "F"
        
        for j = 1:length(testF)             % Compare the 'pruned' sequences to the original ones
            if (nF(testF(j)) == nO(i))      % If sequences are the same length...
                distF(j) = sum(abs(samp_holderO(i,:) - samp_holderF(testF(j),:)));
            else            % else we need to align the sequences 
                distF(j) = align(samp_holderO(i,:), samp_holderF(testF(j),:), n , N);
            end
        end
        
        distF_sort = sort(distF);
        
        distanceF(i) = distF_sort(K);       % Distance to the Kth nearest neighbor
        
        if (distanceF(i) < N*(gapF + 1))
            doneF = 1;          % We only looked at a subset of the change point solutions.
                % Since the gap penalty is N, we only have to expand our
                % search if the Kth distance is at least as large as the
                % gap penalty.
        else
            disp('Expand search')       % A marker to indicate how many times we had to expand our search
            gapF = gapF +1;
            
        end
        
    end
    
    % Now we repeat the entire process except using the "O" solutions
    while (~doneO)
        
        enough = 0;
        while (~enough)
            testO = find(nO >= nO(i)-gapO & nO <= nO(i)+gapO);       % These are all the samples of the same length
            if (length(testO) > K)
                enough = 1;                             % We potentially have enough samples
            else
                gapO = gapO + 1;                          % Otherwise let's expand the # chgpts we will consider
            end
        end
        
        distO = zeros(length(testO), 1);    % Holds distance to each possible change point solution identified in set "O"
        
        for j = 1:length(testO)         % Compare the original sequences to themselves
            if (nO(testO(j)) == nO(i))  % If the sequences are the same length...
                distO(j) = sum(abs(samp_holderO(i,:) - samp_holderO(testO(j),:)));
            else        % else we have to align them before calculating a distance
                distO(j) = align(samp_holderO(i,:), samp_holderO(testO(j),:), n , N);
            end
        end
        
        distO_sort = sort(distO);
        
        distanceO(i) = distO_sort(K+1);         % Since the original set will always compare to itself one time
        
        if (distanceO(i) < N*(gapO + 1))
            doneO = 1;      % We only looked at a subset of the change point solutions.
                % Since the gap penalty is N, we only have to expand our
                % search if the Kth distance is at least as large as the
                % gap penalty.
        else
            disp('Expand Search')   % A marker to indicate how many times we had to expand our search
            gapO = gapO +1;
           
        end
        
    end
    
    d(i) = nO(i) + gapO;        % Determines the dimension of each solution.  
                                % Decision to embed all solutions in the largest dimension being considered.  
end

temp = 0;
for i = 1:num_samp
    temp = temp + d(i)*log(distanceF(i)/distanceO(i));
end

measure = log( num_samp/(num_samp-1) ) + temp/num_samp
% Allows each solution to be in a different dimension
measure2 = log( num_samp/(num_samp-1) ) + n/num_samp*sum(log(distanceF./distanceO))
% Gives all solutions the largest possible dimension (kmax)

function dist = align(chgpt1, chgpt2, n, N)
% Alignment algorithm.  The distance between two change point solutions is
% the bottom right corner of the table being constructed.  GAp penalty is
% set to the length of the data set.

        n1 = min([find(chgpt1 ==0,1)-1 n]);
        n2 = min([find(chgpt2 ==0,1)-1 n]);
        
        align1 = zeros(n1+1,n2+1);     % <-- size of table should be number of non-zero entries
       
        %l = 1;
        for m = 2:n2+1
            align1(1,m) = N + align1(1,m-1);
        end
        
        %m = 1;
        for l = 2:n1+1
            align1(l,1) = align1(l-1,1)+N;
        end
        
        for l = 2:n1+1
            for m = 2:n2+1
                TEMP = [align1(l-1,m)+N; align1(l-1,m-1)+ abs(chgpt1(l-1) - chgpt2(m-1)); align1(l,m-1)+N]';
                align1(l,m) = min(TEMP);
            end
        end
        
        dist = align1(n1+1,n2+1);
    
end

end
